################################################################################
# Probability Class                                                                   #
#                                                                              #
"""Class defines a probability object."""
################################################################################


class P:
    """This class defines a probability object.

    Parameters
    ----------
    p : dictionary, float
        Probability each hour each weekday, either a float for the same
        probability each hour, dictionary of hours for same hour distribution
        for all days, a dictionary of days for the same hour probability for
        different days, or a dictionary of days each with a dictionary of hours
    """
    def __init__(self, p):
        # Process input
        self.set_p(p)


    ##################
    # Setter Methods #
    ##################
    def set_p(self, p):
        """Set complete probability dictionary.

        Parameters
        ----------
        p : dictionary, float
            Probability each hour each weekday, either a float for the same
            probability each hour, dictionary of hours for same hour
            distribution for all days, a dictionary of days for the same hour
            probability for different days, or a dictionary of days each with a
            dictionary of hours
        """
        if p:
            if isinstance(p, float) or isinstance(p, int):
                self._p = {day: {hour: p for hour in range(24)} for day in range(7)}
            elif isinstance(p, dict):
                if len(p.keys())==7 and 0 in p.keys() and isinstance(p[0], dict) and len(p[0].keys())==24:
                    self._p = p
                elif 23 in p.keys():
                    self._p = {day: {hour: p[hour] for hour in range(24)} for day in range(7)}
                elif 6 in p.keys():
                    self._p = {day: {hour: p[day] for hour in range(24)} for day in range(7)}
                else:
                    print("P: Invalid dictionary input...")
                    self._p = None
            else:
                print("P: Invalid probability input...")
                self._p = None
        else:
            print("P: Dictionary p cannot be empty...")
            self._p = None


    def set_p_day(self, day, val):
        """Set probability for a day.

        Parameters
        ----------
        day : integer
            Day index
        val : dictionary
            Probability dictionary for a day with each hour
        """
        self._p[day] = val

    def set_p_hour(self, day, hour, val):
        """Set probability for an hour.

        Parameters
        ----------
        day : integer
            Day index
        hour : integer
            Hour index
        val : float
            Probability of selected hour
        """
        self._p[day][hour] = val


    ##################
    # Getter Methods #
    ##################
    def get_p(self):
        """Return complete probability dictionary.

        Returns
        -------
        val : dictionary
            Probability dictionary for each day and hour
        """
        return self._p

    def get_p_day(self, day):
        """Return probability for a day.

        Parameters
        ----------
        day : integer
            Day index

        Returns
        -------
        val : dictionary
            Probability dictionary for a day with each hour
        """
        return self._p[day]

    def get_p_hour(self, day, hour):
        """Return probability for an hour.

        Parameters
        ----------
        day : integer
            Day index
        hour : integer
            Hour index

        Returns
        -------
        val : float
            Probability of selected hour
        """
        return self._p[day][hour]



from chargesim.probability import P
from chargesim.user import User
from chargesim.car import Car
from chargesim.poi import Poi
from chargesim.topology import Topology
from chargesim.mc import MC

import chargesim.utils as utils

__all__ = [
    "P", "User", "Car", "Poi",
    "Topology", "MC",
    "utils"
]



################################################################################
# Poi Class                                                                    #
#                                                                              #
"""Class defining points of interest."""
################################################################################


from chargesim.probability import P


class Poi(P):
    """This class defines a point of interest.

    Parameters
    ----------
    topo : Topology
        Topology object
    name : string
        Name of point of interest
    p : dictionary, float
        Probability each hour each weekday, either a float for the same
        probability each hour, dictionary of hours for same hour distribution
        for all days, a dictionary of days for the same hour probability for
        different days, or a dictionary of days each with a dictionary of hours
    radius : float, optional
        Distance of nodes from poi center in m
    """
    def __init__(self, topo, name, p, radius=200):
        # Call super class
        super(Poi, self).__init__(p)

        # Process input
        self._topo = topo
        self._name = name

        # Load graph
        self._G = topo.poi(name, radius=radius)
        self._nodes = list(self._G)


    ##################
    # Getter Methods #
    ##################
    def get_topo(self):
        """Return Topology object.

        Returns
        -------
        val : Topology
            Topology object
        """
        return self._topo

    def get_name(self):
        """Return poi name.

        Returns
        -------
        val : string
            Poi name
        """
        return self._name

    def get_G(self):
        """Get OSM graph object.

        Returns
        -------
        val : networkx.MultiDiGraph
            OSM graph object
        """
        return self._G

    def get_nodes(self):
        """Get list of nodes of graph.

        Returns
        -------
        val : list
            List of node ids of graph
        """
        return self._nodes



################################################################################
# MC Class                                                                     #
#                                                                              #
"""Monte Carlo algorithm for calculating occupancy probability."""
################################################################################


import sys
import math
import random
import multiprocessing as mp

from chargesim.probability import P


class MC:
    """This class run a Monte Carlo simulation. Probability of every node is the
    sum of given probabilities from the poi list. This means, nodes intersecting
    multiple poi.s have a higher probability than the single pois.

    Parameters
    ----------
    topo : Topology
        Topology object
    users : list
        List of user objects
    pois : list, optional
        List of poi objects
    node_p : dictionary, float, optional
        Probability of all nodes not covered by pois each hour each weekday,
        either a float for the same probability each hour, dictionary of hours
        for same hour distribution for all days, a dictionary of days for the
        same hour probability for different days, or a dictionary of days each
        with a dictionary of hours
    """
    def __init__(self, topo, users, pois=[], node_p=1):
        # Initialize
        self._topo = topo
        self._users = users
        node_p = P(node_p)

        # Process pois
        p = {}
        # max_p = 0
        for poi in pois:
            for node in poi.get_nodes():
                if node not in p:
                    p[node] = P(p=poi.get_p())
                else:
                    temp_p = {}
                    for day in range(7):
                        temp_p[day] = {}
                        for hour in range(24):
                            temp_p[day][hour] = p[node].get_p_hour(day, hour)+poi.get_p_hour(day, hour)
                            # max_p = temp_p[day][hour] if temp_p[day][hour]>max_p else max_p
                    p[node] = P(p=temp_p)

        # Normalize if probability is greater than 1
        # if max_p > 1:
        #     for node in p:
        #         p[node] = P(p={day: {hour: p[node].get_p_hour(day, hour)/max_p for hour in range(24)} for day in range(7)})

        # Fill empty nodes
        for node in topo.get_nodes():
            if node not in p:
                p[node] = node_p

        self._nodes = p

        # Get charging stations
        self._charge_G, self._capacity = topo.charging_station()
        self._stations = {station: 0 for station in self._capacity.keys()}

    def run(self, weeks, drivers, np=1):
        """Run Monte Carlo code.

        Parameters
        ----------
        weeks : integer
            Number of weeks to simulate
        drivers : dictionary, integer
            Number of drivers each hour, either an integer for the same hour,
            dictionary of hours or dictionary of days with a dictionary of hours
        np : integer, optional
            Number of cores to use, set zero to use all available

        Returns
        -------
        results : dictionary
            Node occupance
        """
        # Initialize
        progress_form = "%"+str(len(str(weeks*7)))+"i"

        # Process input
        if isinstance(drivers, int):
            drivers = {day: {hour: drivers for hour in range(24)} for day in range(7)}
        elif isinstance(drivers, dict):
            if not (0 in drivers.keys() and isinstance(drivers[0], dict)):
                if 23 in drivers.keys():
                    drivers = {day: {hour: drivers[hour] for hour in range(24)} for day in range(7)}
                elif 6 in drivers.keys():
                    drivers = {day: {hour: drivers[day] for hour in range(24)} for day in range(7)}
                else:
                    print("MC: Invalid dirver input...")
                    return
        else:
            print("MC: Invalid dirver input...")
            return

        # Get number of cores
        np = np if np and np<=mp.cpu_count() else mp.cpu_count()

        # Run through weeks
        for week in range(weeks):
            # Run through days
            for day in range(7):
                # Run through hours
                for hour in range(24):
                    # Run through dirvers
                    if np>1:
                        # # Distribute drivers on processors
                        # driver_num = math.floor(drivers[day][hour]/np)
                        # driver_np = [drivers[day][hour]-driver_num*(np-1) if i == np-1 else driver_num for i in range(np)]
                        #
                        # # Run parallelization
                        # pool = mp.Pool(processes=np)
                        # results = [pool.apply_async(self._run_helper, args=(x, day, hour,)) for x in driver_np]
                        # pool.close()
                        # pool.join()
                        # output = [x.get() for x in results]
                        #
                        # # Destroy object
                        # del results

                        output = [self._run_helper(drivers[day][hour], day, hour)]
                    else:
                        # Run sampling
                        output = [self._run_helper(drivers[day][hour], day, hour)]

                    # Add stations
                    for stations  in output:
                        for station, capacity in stations.items():
                            self._stations[station] += capacity

                # Progress
                sys.stdout.write("Finished day "+progress_form%(week*7+day+1)+"/"+progress_form%(weeks*7)+"...\r")
                sys.stdout.flush()
        print()

    def _run_helper(self, drivers, day, hour):
        """Helper function for parallelization. This function goes through a
        number of drivers and calculates the nodes and walking distances to the
        charging stations.

        Parameters
        ----------
        drivers : integer
            Number of drivers to process
        day : integer
            Day number
        hour : integer
            Hour number

        Returns
        -------
        stations : dictionary
            Dictionary containing all destinations and their number of entries
        """
        # Initialize
        stations = {}

        # Run through dirvers
        for driver in range(drivers):
            # Choose random user
            user = random.choice(self._users)
            rand = random.uniform(0, 1)
            # User MC step
            if rand <= user.get_p_hour(day, hour):
                # Choose random node
                node = random.choice(list(self._nodes.keys()))
                rand = random.uniform(0, 1)
                # Poi MC step
                if rand <= self._nodes[node].get_p_hour(day, hour):
                    # Calculate distance to nearest charging station
                    dest, dist = self._topo.dist_poi(node, self._charge_G)

                    # Add stations
                    if dest in stations:
                        stations[dest] += 1
                    else:
                        stations[dest] = 1

        return stations



import os
import sys

import shutil
import unittest

import osmnx as ox
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

import chargesim as cs


class UserModelCase(unittest.TestCase):
    #################
    # Remove Output #
    #################
    @classmethod
    def setUpClass(self):
        if os.path.isdir("tests"):
            os.chdir("tests")

        folder = 'output'
        cs.utils.mkdirp(folder)
        cs.utils.mkdirp(folder+"/temp")
        open(folder+"/temp.txt", 'a').close()

        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)

        # Set style
        sns.set_style("white",{"xtick.bottom": True,'ytick.left': True})
        sns.set_context("paper")
        sns.set_palette(sns.color_palette("deep"))


    #########
    # Utils #
    #########
    def test_utils(self):
        file_link = "output/test/test.txt"

        cs.utils.mkdirp("output/test")

        with open(file_link, "w") as file_out:
            file_out.write("TEST")
        cs.utils.copy(file_link, file_link+"t")
        cs.utils.replace(file_link+"t", "TEST", "DOTA")
        with open(file_link+"t", "r") as file_in:
            for line in file_in:
                self.assertEqual(line, "DOTA\n")

        self.assertEqual(cs.utils.column([[1, 1, 1], [2, 2, 2]]), [[1, 2], [1, 2], [1, 2]])

        cs.utils.save([1, 1, 1], file_link)
        self.assertEqual(cs.utils.load(file_link), [1, 1, 1])

        print()
        cs.utils.toc(cs.utils.tic(), message="Test", is_print=True)
        self.assertEqual(round(cs.utils.toc(cs.utils.tic(), is_print=True)), 0)


    ###############
    # Probability #
    ###############
    def test_p(self):
        # Initialize
        p = cs.P(1)
        p = cs.P({day: 1 for day in range(7)})
        p = cs.P({hour: 1 for hour in range(24)})
        p = cs.P({day: {hour: 1 for hour in range(24)} for day in range(7)})

        # Probability
        p.set_p({day: {hour: 1/7 for hour in range(24)} for day in range(7)})
        p.set_p_day(0, {0: 1, 1: 1, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1})
        self.assertEqual(p.get_p_day(0), {0: 1, 1: 1, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1})
        p.set_p_hour(0, 0, 0.5)
        self.assertEqual(p.get_p_hour(0, 0), 0.5)
        self.assertEqual(p.get_p()[0][0], 0.5)

        # Check errors
        print()
        self.assertIsNone(cs.P({}).get_p())
        self.assertIsNone(cs.P("DOTA").get_p())
        self.assertIsNone(cs.P({0: 1}).get_p())


    ########
    # User #
    ########
    def test_user(self):
        # Initialize
        user = cs.User(1)

        # Ident
        user.set_ident("DOTA")
        self.assertEqual(user.get_ident(), "DOTA")


    #######
    # Car #
    #######
    def test_car(self):
        # Initialize
        car = cs.Car(100)

        # Ident
        car.set_ident("DOTA")
        self.assertEqual(car.get_ident(), "DOTA")

        # Size
        car.set_size(150)
        self.assertEqual(car.get_size(), 150)


    ############
    # Topology #
    ############
    def test_topo(self):
        # self.skipTest("Temporary")

        # Load
        name = "Munich, Bavaria, Germany"
        # name = "Piedmont, California, USA"
        # topo = cs.Topology({"name": name}, tags={})
        # cs.utils.save(topo.get_G(), "data/munich_G.obj")
        # cs.utils.save(topo.get_Gp(), "data/munich_Gp.obj")

        G = cs.utils.load("data/munich_G.obj")
        Gp = cs.utils.load("data/munich_Gp.obj")
        topo = cs.Topology({"name": name, "G": G, "Gp": Gp})
        self.assertEqual(list(topo.get_G())[0], 128236)
        self.assertEqual(list(topo.get_Gp())[0], 128236)
        self.assertEqual(topo.get_nodes()[0], 128236)

        # Poi
        P, _ = topo.poi("cafe", radius=0, is_gdf=True)
        P = topo.poi("cafe", radius=300)

        # Charging stations
        C, capacity = topo.charging_station()
        self.assertEqual(list(C)[0], 5156040713)
        self.assertEqual(capacity[1249710076], 4)

        # Distance
        dest, route_len = topo.dist_poi(1955541, C)
        self.assertEqual(round(route_len, 2), 333.71)
        route_len, route = topo.dist(1955541, dest, True)

        # Plot
        topo.plot(pois=[P])
        plt.savefig("output/topo_cafe.pdf", format="pdf", dpi=1000)

        topo.plot(routes=[route, route])
        topo.plot(routes=[route])
        topo.plot(pois=[C])
        plt.savefig("output/topo_charge.pdf", format="pdf", dpi=1000)


    #######
    # Poi #
    #######
    def test_poi(self):
        # self.skipTest("Temporary")

        # Initialize
        name = "Munich, Bavaria, Germany"
        G = cs.utils.load("data/munich_G.obj")
        Gp = cs.utils.load("data/munich_Gp.obj")
        topo = cs.Topology({"name": name, "G": G, "Gp": Gp})
        poi = cs.Poi(topo, "cafe", 1)

        self.assertEqual(poi.get_topo(), topo)
        self.assertEqual(poi.get_name(), "cafe")
        self.assertEqual(list(poi.get_G())[0], 3571318797)
        self.assertEqual(poi.get_nodes()[0], 3571318797)


    ######
    # MC #
    ######
    def test_mc(self):
        # self.skipTest("Temporary")

        # Topology
        name = "Munich, Bavaria, Germany"
        G = cs.utils.load("data/munich_G.obj")
        Gp = cs.utils.load("data/munich_Gp.obj")
        topo = cs.Topology({"name": name, "G": G, "Gp": Gp}, is_log=False)

        # Poi
        pois = [cs.Poi(topo, "cafe", 0.3), cs.Poi(topo, "restaurant", 0.3), cs.Poi(topo, "bar", 0.3)]

        # User
        users = [cs.User(1)]

        # Initialize
        mc = cs.MC(topo, users, pois, node_p=0.1)
        # mc = cs.MC(topo, users, node_p=1)

        # Run MC
        mc.run(1, {day: 0 for day in range(7)})
        mc.run(1, {hour: 0 for hour in range(24)})
        mc.run(1, {day: {hour: 0 for hour in range(24)} for day in range(7)})
        mc.run(1, 20, np=0)

        # Plot occupancy
        occ = mc._stations
        data = [{"node": node, "x": G.nodes[node]["x"], "y": G.nodes[node]["y"], "capacity": capacity} for node, capacity in occ.items()]
        df = pd.DataFrame(data)
        print(df)

        plt.figure(figsize=(6, 4))
        sns.scatterplot(data=df, x="x", y="y", hue="capacity", size="capacity",
            palette="ch:r=-.2,d=.3_r", sizes=(1, 8), linewidth=0)
        plt.savefig("output/mc.pdf", format="pdf", dpi=1000)

        # Check errors
        self.assertIsNone(mc.run(1, "DOTA"))
        self.assertIsNone(mc.run(1, {0: 1}))


if __name__ == '__main__':
    unittest.main(verbosity=2)



################################################################################
# Car Class                                                                    #
#                                                                              #
"""Class defining car types."""
################################################################################


class Car:
    """This class defines a car.

    Parameters
    ----------
    size : float
        Battery size in kWh
    ident : string, optional
        Optional car name
    """
    def __init__(self, size, ident=""):
        # Process input
        self._size = size
        self._ident = ident


    ##################
    # Setter Methods #
    ##################
    def set_size(self, val):
        """Set battery size.

        Parameters
        ----------
        val : float
            Battery size in kWh
        """
        self._size = val

    def set_ident(self, val):
        """Set car name.

        Parameters
        ----------
        val : string
            Car name
        """
        self._ident = val


    ##################
    # Getter Methods #
    ##################
    def get_size(self):
        """Return battery size.

        Returns
        -------
        val : float
            battery size in kWh
        """
        return self._size

    def get_ident(self):
        """Return car name.

        Returns
        -------
        val : string
            Car name
        """
        return self._ident



################################################################################
# Topology Class                                                               #
#                                                                              #
"""OSMNX python wrapper."""
################################################################################


import math
import osmnx as ox
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt


class Topology:
    """This class is a python wrapper for the osmnx package. The input is a
    dictionary containing the location name **name**, and optionally preloaded
    graph **G** and projection objects **Gp**.

    Parameters
    ----------
    loc : dictionary
        Location for simulation as dictionary {"name": "", "G": None, "Gp": None}
    is_log : bool, optional
        True to print osmnx console output
    """
    def __init__(self, loc, is_log=True):
        # Set osmnx output
        ox.config(log_console=is_log)
        self._loc = loc

        # Process input
        self._G = ox.graph_from_place(loc["name"], network_type="walk") if not "G" in loc.keys() else loc["G"]
        self._Gp = ox.project_graph(self._G) if not "Gp" in loc.keys() else loc["Gp"]
        self._nodes = list(self._G)

    def dist(self, orig, dest, is_route=False):
        """Calculate the distance between two locations

        Parameters
        ----------
        orig : integer
            Node of origin
        dest : integer
            Node of destination
        is_route : bool
            True to return route

        Returns
        -------
        route_len : float
            Route length in m
        route : list, optional
            Route as list of nodes
        """
        # Calculate shortest distance
        route_len = nx.shortest_path_length(self._G, orig, dest, weight="length")

        # Get shortest route
        if is_route:
            route = ox.shortest_path(self._G, orig, dest, weight="length")
            return route_len, route
        else:
            return route_len

    def dist_poi(self, orig, poi):
        """Calculate the distance to nearest poi node.

        Parameters
        ----------
        orig : integer
            Node of origin
        poi : networkx.MultiDiGraph
            Poi graph

        Returns
        -------
        dest : integer
            Node of destination
        route_len : float
            Route length in m
        """
        # Get position of origin
        pos = dict(self._G.nodes(data=True))[orig]

        # Get nearest charging station
        dest = ox.nearest_nodes(poi, pos["x"], pos["y"])

        # Get shortest route
        return dest, self.dist(orig, dest, is_route=False)

    def poi(self, name, radius=0, is_gdf=False):
        """Get nodes for given point of interest.

        Parameters
        ----------
        name : string
            OSM name for point of interest
        radius : float, optional
            Distance of pois from center in m
        is_gdf : bool, optional
            True to return GDF object

        Returns
        -------
        poi : networkx.MultiDiGraph
            OSM graph object
        gdf : geopandas.GeoDataFrame, optional
            GDF object
        """
        # Get pois geometry
        gdf = ox.geometries_from_place(self._loc["name"], tags={"amenity" : [name]})

        # Find nearest nodes in graph
        pos = gdf["geometry"]["node"]
        poi_nodes = ox.nearest_nodes(self._G, pos.x, pos.y)

        # Find nodes within radius
        if radius:
            nodes = []
            for node in poi_nodes:
                poi = nx.ego_graph(self._G, node, radius, distance="length")
                nodes += list(poi)
            nodes = list(set(nodes))
        else:
            nodes = poi_nodes

        # Create subgraph from nodes
        P = self._G.subgraph(nodes)

        # Return
        if is_gdf:
            return P, gdf
        else:
            return P

    def charging_station(self, name="charging_station"):
        """Create charging station graph and capacity dictionaries.

        Parameters
        ----------
        name : string, optional
            Charging station tag name

        Returns
        -------
        stations : networkx.MultiDiGraph
            OSM graph object containing all charging stations
        capacity : dictionary
            Charging stations capacities
        """
        # Get chargin stations
        gdf = ox.geometries_from_place(self._loc["name"], tags={"amenity" : [name]})

        # Find nearest nodes in graph
        pos = gdf["geometry"]
        nodes = ox.nearest_nodes(self._G, pos.x, pos.y)

        # Create subgraph from nodes
        C = self._G.subgraph(nodes)

        # Create list conatining capacities
        capacity_list = gdf["capacity"]
        capacity_list = pd.to_numeric(capacity_list, downcast="integer")
        capacity_list = {node: int(val) if not math.isnan(val) else 1 for node, val in dict(capacity_list["node"]).items()}

        capacity = {}
        for i, node in enumerate(pos["node"].index):
            node_new = nodes[i]
            if node_new not in capacity:
                capacity[node_new] = capacity_list[node]
            else:
                capacity[node_new] += capacity_list[node]

        return C, capacity

    def plot(self, pois=[], routes=[], ax=None, kwargs={"G": {}, "P": {}, "R": {}}):
        """Plot graph optionally with chargin stations and routes.

        Parameters
        ----------
        poi : list, optional
            List of points of interest graphs to show
        routes : list, optional
            List of routes to show
        ax : Axis, optional
            Axis object
        kwargs: dictionary, optional
            Dictionary with plotting parameters for the graph **G**, Pois **P**
            and routes **R**

        Returns
        -------
        ax : Axis
            Axis object
        """
        # Initialize
        if ax is None:
            _, ax = plt.subplots(figsize=(12, 8))

        # Plot edges
        ox.plot_graph(self._G, ax=ax, node_size= 0, edge_linewidth=1, edge_color="#262626", show=False, edge_alpha=0.1, **kwargs["G"])

        # Plot charging stations
        if pois:
            for poi in pois:
                ox.plot_graph(poi, ax=ax, node_size=40, edge_linewidth=0, node_color="#4C72B0", show=False, **kwargs["P"])

        # Plot routes
        if routes:
            if len(routes)==1:
                ox.plot_graph_route(self._G, routes[0], ax=ax, route_color="#C44E52", route_linewidth=6, node_size=0, **kwargs["R"])
            else:
                ox.plot_graph_routes(self._G, routes, ax=ax, route_color="#C44E52", route_linewidth=6, node_size=0, **kwargs["R"])

        return ax


    ##################
    # Getter Methods #
    ##################
    def get_G(self):
        """Get OSM graph object.

        Returns
        -------
        val : networkx.MultiDiGraph
            OSM graph object
        """
        return self._G

    def get_Gp(self):
        """Get OSM graph projection object.

        Returns
        -------
        val : networkx.MultiDiGraph
            OSM graph projection object
        """
        return self._Gp

    def get_nodes(self):
        """Get list of nodes of graph.

        Returns
        -------
        val : list
            List of node ids of graph
        """
        return self._nodes



################################################################################
# Utils                                                                        #
#                                                                              #
"""Here popular basic methods are noted."""
################################################################################


import os
import time
import pickle
import fileinput

from shutil import copyfile


def mkdirp(directory):
    """Create directory if it does not exist.

    Parameters
    ----------
    directory : string
        Directory name
    """
    if not os.path.exists(directory):
        os.makedirs(directory)


def copy(source, target):
    """Copy a specified file to a specified location.

    Parameters
    ----------
    source : string
        Link to requested file
    target : string
        Link to requested new file
    """
    copyfile(source, target)


def column(data):
    """Convert given row list matrix into column list matrix

    Parameters
    ----------
    data : list
        Row data matrix

    Returns
    -------
    data_col : list
        column data matrix
    """
    num_row = len(data)
    num_col = len(data[0])

    data_col = [[] for i in range(num_col)]

    for i in range(num_row):
        for j in range(num_col):
            data_col[j].append(data[i][j])

    return data_col


def tic():
    """MATLAB tic replica - return current time.

    Returns
    -------
    time : float
        Current time in seconds
    """
    return time.time()


def toc(t, message="", is_print=True):
    """MATLAB toc replica - return time difference to tic and alternatively
    print a message.

    Parameters
    ----------
    t : float
        Starting time - given by tic function
    message : string, optional
        Custom output message
    is_print : bool, optional
        True for printing an output message

    Returns
    -------
    time : float
        Time difference
    """
    if message:
        message += " - runtime = "

    t_diff = time.time()-t

    if is_print:
        print(message+"%6.3f" % t_diff+" s")

    return t_diff


def replace(file_link, old, new):
    """Replace a given string in a file.

    Parameters
    ----------
    file_link : string
        Link to requested file
    old : string
        String to be replaced in file
    new : string
        New string to be written
    """
    for line in fileinput.input(file_link, inplace=True):
        print(line.rstrip().replace(old, new))


def save(obj, link):
    """Save an object using pickle in the specified link.

    Parameters
    ----------
    obj : Object
        Object to be saved
    link : string
        Specific link to save object
    """
    with open(link, "wb") as f:
        pickle.dump(obj, f)


def load(link):
    """Load pickled object from the specified folder.

    Parameters
    ----------
    link : string
        Specific link to load object

    Returns
    -------
    obj : Object
        Loaded object
    """
    with open(link, 'rb') as f:
        return pickle.load(f)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="chargesim",
    version="0.0.1",
    author="Hamzeh Kraus",
    author_email="hamzeh_kraus@web.de",
    description="Charging infrastructure simulation suite.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ajax23/ChargeSim",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    install_requires=['osmnx', 'pandas', 'seaborn', 'scikit-learn'],
    include_package_data=True,
)



################################################################################
# User Class                                                                   #
#                                                                              #
"""Class defining user types."""
################################################################################


from chargesim.probability import P


class User(P):
    """This class defines a user.

    Parameters
    ----------
    p : dictionary, float
        Probability each hour each weekday, either a float for the same
        probability each hour, dictionary of hours for same hour distribution
        for all days, a dictionary of days for the same hour probability for
        different days, or a dictionary of days each with a dictionary of hours
    ident : string, optional
        Optional user name
    """
    def __init__(self, p, ident=""):
        # Call super class
        super(User, self).__init__(p)

        # Process input
        self._ident = ident


    ##################
    # Setter Methods #
    ##################
    def set_ident(self, val):
        """Set user name.

        Parameters
        ----------
        val : string
            User name
        """
        self._ident = val


    ##################
    # Getter Methods #
    ##################
    def get_ident(self):
        """Return user name.

        Returns
        -------
        val : string
            User name
        """
        return self._ident
